/*
 *  This file is part of the nuvtools package
 *  Copyright (C) 2004, 2005 Gavin Hurlbut
 *
 *  nuvtools is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*HEADER---------------------------------------------------
* $Id: encode.c 105 2005-04-24 03:10:16Z gjhurlbu $
*
* Copyright 2004, 2005 Gavin Hurlbut
* All rights reserved
*
*/

#include "environment.h"
#include <stdlib.h>
#include "format.h"
#include "nuvread.h"
#include <lame/lame.h>
#include "avcodec.h"
#include "types.h"
#include "logging.h"

/* CVS generated ID string */
static char ident[] _UNUSED_ = 
    "$Id: encode.c 105 2005-04-24 03:10:16Z gjhurlbu $";


static lame_global_flags *gf;
static int              lameret;
static AVCodecContext  *avOutCtx;
static int              ffmpegInit = FALSE;

void initAudioEncode( nuvFile *nv )
{
    int             tmp;

    gf = lame_init();

    lame_set_in_samplerate(gf, nv->audio_sample_rate);
    lame_set_num_channels(gf, nv->audio_channels);
    lame_set_out_samplerate(gf, nv->audio_sample_rate);
    lame_set_bWriteVbrTag(gf, 0);
    lame_set_quality(gf, nv->audio_quality);
    lame_set_mode(gf, nv->audio_channels == 2 ? STEREO : MONO);
    lame_set_brate(gf, nv->audio_bitrate_closest / 1000);

    if( (tmp = lame_init_params(gf)) != 0 ) {
        LogPrint( LOG_CRIT, "LAME initialization failed: %d\n", tmp );
    }
    lame_print_config(gf);
}

nuvFrame *encodeAudio( nuvFile *nv, int interleaved, short int *pcmL,
                       short int *pcmR, int pcmSize )
{
    int             count;
    int             mp3size;
    nuvFrame       *frame;

    frame = nuv_new_frame();
    if( !frame ) {
        return( NULL );
    }

    frame->header.frametype = 'A';

    if( interleaved ) {
        count = pcmSize / (nv->audio_channels * sizeof(short int));
    } else {
        count = pcmSize / sizeof(short int);
    }

    mp3size = ((count * 5) / 4) + 14400;
    frame->data = (char *)malloc(mp3size);
    if( !frame->data ) {
        fprintf( stderr, "Couldn't allocate a frame, aborting!\n" );
        free( frame );
        return( NULL );
    }

    if( interleaved ) {
        frame->length = lame_encode_buffer_interleaved(gf, pcmL, count, 
                                                       frame->data, mp3size);
    } else {
        frame->length = lame_encode_buffer(gf, pcmL, pcmR, count, frame->data,
                                           mp3size);
    }

    if( frame->length < 0 ) {
        frame->length = 0;
    }

    frame->length += lame_encode_flush_nogap(gf, &(frame->data[frame->length]),
                                             7200);
    if( frame->length < 0 ) {
        frame->length = 0;
    }

    return( frame );
}

void closeAudio( void )
{
    lame_close(gf);
}


void initVideo( nuvFile *nv )
{
    AVCodec *codec;

    codec = avcodec_find_encoder_by_name("mpeg4");
    if( !codec ) {
        fprintf( stderr, "Couldn't find mpeg4 codec!\n");
        exit( -1 );
    }

    avOutCtx = avcodec_alloc_context();
    if( !avOutCtx ) {
        fprintf( stderr, "Couldn't initialize ffmpeg context!  Aborting!\n" );
        exit(-1);
    }

    avOutCtx->pix_fmt = PIX_FMT_YUV420P;
    avOutCtx->width = nv->width;
    avOutCtx->height = nv->height;
    avOutCtx->frame_rate = nv->fps.num;
    avOutCtx->frame_rate_base = nv->fps.denom;
    avOutCtx->bit_rate = 2000000;
    avOutCtx->bit_rate_tolerance = (int)(avOutCtx->bit_rate * 0.2);
    avOutCtx->qmin = avOutCtx->mb_qmin = 5;
    avOutCtx->qmax = avOutCtx->mb_qmax = 31;
    avOutCtx->max_qdiff = 10;
    avOutCtx->flags = 0;
    avOutCtx->mb_decision = 0;
    avOutCtx->qblur = 0.5;
    avOutCtx->max_b_frames = 0;
    avOutCtx->b_quant_factor = 0;
    avOutCtx->rc_strategy = 2;
    avOutCtx->b_frame_strategy = 0;
    avOutCtx->gop_size = 30;
    avOutCtx->rc_max_rate = 0;
    avOutCtx->rc_min_rate = 0;
    avOutCtx->rc_buffer_size = 0;
    avOutCtx->rc_buffer_aggressivity = 1.0;
    avOutCtx->rc_override_count = 0;
    avOutCtx->rc_initial_cplx = 0;
    avOutCtx->dct_algo = FF_DCT_AUTO;
    avOutCtx->idct_algo = FF_IDCT_AUTO;
    avOutCtx->prediction_method = FF_PRED_LEFT;
    
    if( avcodec_open( avOutCtx, codec ) < 0 ) {
        fprintf( stderr, "Couldn't initialize codec!\n" );
        exit(-1);
    }

    ffmpegInit = TRUE;
}

nuvFrame *encodeVideo( nuvFile *nv, nuvFrame *frameIn )
{
    nuvFrame       *vidFrame;
    int             count;
    unsigned char  *planes[3];
    int             frameSize;
    AVFrame        *pict;

    frameSize = nv->width * nv->height;

    planes[0] = frameIn->data;
    planes[1] = planes[0] + frameSize;
    planes[2] = planes[1] + (frameSize / 4);

    /* Create an video frame */
    vidFrame = nuv_new_frame();
    if( !vidFrame ) {
        fprintf( stderr, "Couldn't allocate a frame, aborting!\n" );
        return( NULL );
    }
    vidFrame->header.frametype = 'V';
    vidFrame->header.timecode  = frameIn->header.timecode;

    vidFrame->data = (char *)malloc(nv->frameSize);
    if( !vidFrame->data ) {
        fprintf( stderr, "Couldn't allocate a frame, aborting!\n" );
        return( NULL );
    }

    pict = avcodec_alloc_frame();

    pict->data[0] = planes[0];
    pict->data[1] = planes[1];
    pict->data[2] = planes[2];
    pict->linesize[0] = nv->width;
    pict->linesize[1] = nv->width / 2;
    pict->linesize[2] = nv->width / 2;
    pict->pts = vidFrame->header.timecode * 1000;
    pict->type = FF_BUFFER_TYPE_SHARED;
    if( frameIn->header.keyframe == 0  ) {
        pict->pict_type = FF_I_TYPE;
    } else {
        pict->pict_type = 0;
    }

    if( !ffmpegInit ) {
        initVideo(nv);
    }

    vidFrame->length = avcodec_encode_video( avOutCtx, vidFrame->data,
                                             nv->frameSize, pict );

    av_free(pict);
    nuv_free_frame( frameIn );

    return( vidFrame );
}

void closeVideo( void )
{
    if( ffmpegInit ) {
        avcodec_close( avOutCtx );
        av_free( avOutCtx );
    }
}

/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */
